""" Pydantic models."""

from datetime import datetime
from typing import Optional
from uuid import UUID

from pydantic import BaseModel, Field


class GenerationSum(BaseModel):
    """Sum of generation."""

    power_kw: float = Field(..., description="Summed power in kW")
    start_utc: datetime = Field(..., description="Start datetime of this power")
    name: str = Field(..., description="Name of item sums. ")


class ForecastValueSum(BaseModel):
    """Sum of forecast values."""

    power_kw: float = Field(..., description="Summed power in kW")
    start_utc: datetime = Field(..., description="Start datetime of this power")
    name: str = Field(..., description="Name of item sums. ")


class LatitudeLongitudeLimits(BaseModel):
    """Lat and lon limits."""

    latitude_min: float = Field(-90, description="Minimum latitude")
    latitude_max: float = Field(90, description="Maximum latitude")
    longitude_min: float = Field(-180, description="Minimum longitude")
    longitude_max: float = Field(180, description="Maximum longitude")


class PVSiteEditMetadata(BaseModel):
    """Site metadata when editing a site."""

    client_site_id: Optional[int] = Field(
        None,
        description="The site ID as given by the providing user.",
    )
    client_site_name: Optional[str] = Field(
        None,
        description="The name of the site as given by the providing user.",
    )
    orientation: Optional[float] = Field(
        None,
        description="The rotation of the panel in degrees. 180° points south",
    )
    tilt: Optional[float] = Field(
        None,
        description="The tile of the panel in degrees. 90° indicates the panel is vertical.",
    )
    latitude: Optional[float] = Field(None, description="The site's latitude", ge=-90, le=90)
    longitude: Optional[float] = Field(None, description="The site's longitude", ge=-180, le=180)
    inverter_capacity_kw: Optional[float] = Field(
        None, description="The site's inverter capacity in kw", ge=0
    )
    module_capacity_kw: Optional[float] = Field(
        None,
        description="The site's PV module nameplate capacity in kw",
        ge=0,
    )
    capacity_kw: Optional[float] = Field(None, description="The site's total capacity in kw", ge=0)
    dno: Optional[str] = Field(None, description="The site's DNO")
    gsp: Optional[str] = Field(None, description="The site's GSP")
    client_uuid: Optional[UUID] = Field(
        None, description="The UUID of the client this site belongs to"
    )



""" Functions for reading user data from the database. """

import logging

from sqlalchemy.orm import Session

from pvsite_datamodel.sqlmodels import ClientSQL

logger = logging.getLogger(__name__)


def get_client_by_name(
    session: Session, name: str, make_new_client_if_none: bool = True
) -> ClientSQL:
    """
    Get client by name. If client does not exist, make one.

    :param session: database session
    :param name: name of the client
    :param make_new_client_if_none: make client with name if doesn't exist
    :return: client object
    """

    client = session.query(ClientSQL).filter(ClientSQL.client_name == name).first()

    if client is None:
        if make_new_client_if_none is True:
            logger.info(f"Client with name {name} not found, so making one")

            # make a new client
            client = ClientSQL(client_name=name)
            session.add(client)
            session.commit()
        else:
            raise Exception(f"Could not find client with name {name}")

    return client



"""
Functions to read from the database and format.
"""

import logging

from pvsite_datamodel.read.user import get_user_by_email as get_user_by_db
from pvsite_datamodel.sqlmodels import APIRequestSQL

logger = logging.getLogger(__name__)


def save_api_call_to_db(url, session, user=None):
    """
    Save api call to database.
    """
    url = str(url)
    if user is None:
        email = "unknown"
        user = get_user_by_db(session=session, email=email)

    email = user.email

    # make api call
    logger.info(f"Saving api call ({url=}) to database for user {email}")
    api_request = APIRequestSQL(url=url, user=user)

    # commit to database
    session.add(api_request)
    session.commit()



""" Read functions for getting ML models. """
import logging
from datetime import datetime
from typing import List, Optional

from sqlalchemy.orm import Session

from pvsite_datamodel.sqlmodels import ForecastSQL, ForecastValueSQL, MLModelSQL, SiteSQL

logger = logging.getLogger(__name__)


def get_or_create_model(session: Session, name: str, version: Optional[str] = None) -> MLModelSQL:
    """
    Get model object from name and version.

    A new model is made if it doesn't not exists

    :param session: database session
    :param name: name of the model
    :param version: version of the model

    return: Model object

    """

    # start main query
    query = session.query(MLModelSQL)

    # filter on gsp_id
    query = query.filter(MLModelSQL.name == name)
    if version is not None:
        query = query.filter(MLModelSQL.version == version)

    # gets the latest version
    query = query.order_by(MLModelSQL.version.desc())

    # get all results
    models = query.all()

    if len(models) == 0:
        logger.debug(
            f"Model for name {name} and version {version} does not exist so going to add it"
        )

        model = MLModelSQL(name=name, version=version)
        session.add(model)
        session.commit()

    else:
        model = models[0]

    return model


def get_models(
    session: Session,
    start_datetime: Optional[datetime] = None,
    end_datetime: Optional[datetime] = None,
    site_uuid: Optional[str] = None,
) -> List[MLModelSQL]:
    """
    Get model names from forecast values.

    They are distinct on model name
    By adding start and end datetimes, we only look at forecast values in that time range.
    By adding site_uuid, we only look at forecast values for that site.

    :param session: database session
    :param start_datetime: optional filter on start datetime
    :param end_datetime: optional filter on end datetime
    :param site_uuid: optional filter on site uuid
    :return: list of model names
    """
    query = session.query(MLModelSQL)

    query = query.distinct(MLModelSQL.name)

    if (start_datetime is not None) or (end_datetime is not None) or (site_uuid is not None):
        query = query.join(ForecastValueSQL)

    if start_datetime is not None:
        query = query.where(ForecastValueSQL.start_utc > start_datetime)

    if end_datetime is not None:
        query = query.where(ForecastValueSQL.start_utc < end_datetime)

    if site_uuid is not None:
        query = query.join(ForecastSQL)
        query = query.join(SiteSQL)
        query = query.where(SiteSQL.site_uuid == site_uuid)

    models: [MLModelSQL] = query.all()
    return models



"""Functions for reading to pvsite db."""

import logging
from typing import List, Optional

from sqlalchemy.orm import Session

from pvsite_datamodel.pydantic_models import LatitudeLongitudeLimits
from pvsite_datamodel.sqlmodels import ClientSQL, SiteGroupSiteSQL, SiteGroupSQL, SiteSQL, UserSQL

logger = logging.getLogger(__name__)


def get_site_by_uuid(session: Session, site_uuid: str) -> SiteSQL:
    """Get site object from uuid.

    Raise error if site does not exist.
    :param session: database sessions
    :param site_uuid: the site uuid
    :return: the site object
    """
    query = session.query(SiteSQL)
    query = query.filter(SiteSQL.site_uuid == site_uuid)
    existing_site: Optional[SiteSQL] = query.first()
    if existing_site is None:
        raise KeyError(f"Site uuid {site_uuid} not found in sites table")

    return existing_site


def get_site_by_client_site_id(session: Session, client_name: str, client_site_id: int) -> SiteSQL:
    """Get site from client name and client site id.

    :param session: database sessions
    :param client_name: client name
    :param client_site_id: client's id of site
    :return: site object, or None
    """
    logger.debug(f"Getting {client_name}'s site {client_site_id}")

    # start main query
    query = session.query(SiteSQL)

    # start main query
    query = query.filter(SiteSQL.client_site_name == client_name)

    # select the correct client site id
    query = query.filter(SiteSQL.client_site_id == client_site_id)

    # get first result (should only be one site)
    site: Optional[SiteSQL] = query.first()

    if site is None:
        raise KeyError(f"Could not find site {client_site_id} from client {client_name}")

    return site


def get_site_by_client_site_name(
    session: Session, client_name: str, client_site_name: str
) -> SiteSQL:
    """Get site from client name and client site id.

    :param session: database sessions
    :param client_name: client name
    :param client_site_name: client's name of site
    :return: site object, or None
    """
    logger.debug(f"Getting {client_name}'s site {client_site_name}")

    # start main query
    query = session.query(SiteSQL)

    # select the correct client site name
    query = query.filter(SiteSQL.client_site_name == client_site_name)

    # get first result (should only be one site)
    site: Optional[SiteSQL] = query.first()

    if site is None:
        raise Exception(f"Could not find site {client_site_name} from client {client_name}")

    return site


def get_all_sites(session: Session) -> List[SiteSQL]:
    """Get all sites from the sites table.

    :param session: database session
    :return: site object
    """
    logger.debug("Getting all sites")

    # start main query
    query = session.query(SiteSQL)

    # order by uuuid
    query = query.order_by(SiteSQL.site_uuid)

    # get all results
    sites = query.all()

    logger.debug(f"Found {len(sites)} sites")

    return sites


def get_sites_by_country(
    session: Session, country: str, client_name: Optional[str] = None
) -> List[SiteSQL]:
    """Get sites for specific country from the sites table.

    :param session: database session
    :param country: country name
    :param client_name: optional client string, we search this in the 'client_site_name'.
    :return: site object
    """
    logger.debug(f"Getting sites by country={country}")

    # start main query
    query = session.query(SiteSQL)

    # filter by country
    query = query.filter(SiteSQL.country == country)

    # filter by client string
    # This could cause an issue if client_a's name is in the site of clients_b.
    # We could make a new Client table and join it with sites
    # https://github.com/openclimatefix/pv-site-datamodel/issues/148
    if client_name is not None:
        query = query.filter(SiteSQL.client_site_name.like(f"%{client_name}%"))

    # order by uuuid
    query = query.order_by(SiteSQL.site_uuid)

    # get all results
    sites = query.all()

    logger.debug(f"Found {len(sites)} sites")

    return sites


def get_sites_from_user(session, user, lat_lon_limits: Optional[LatitudeLongitudeLimits] = None):
    """
    Get the sites for a user.

    Option to filter on latitude longitude max and min
    """

    # get sites and filter if required
    if lat_lon_limits is not None:
        # make query
        query = session.query(SiteSQL)

        # join to Usersql through site groups
        query = query.join(SiteGroupSiteSQL)
        query = query.join(SiteGroupSQL)
        query = query.join(UserSQL)

        # filter on user on lat lon limits
        query = query.filter(SiteSQL.latitude <= lat_lon_limits.latitude_max)
        query = query.filter(SiteSQL.latitude >= lat_lon_limits.latitude_min)
        query = query.filter(SiteSQL.longitude <= lat_lon_limits.longitude_max)
        query = query.filter(SiteSQL.longitude >= lat_lon_limits.longitude_min)

        # query db
        sites = query.all()

    else:
        sites = user.site_group.sites
    return sites


def get_sites_by_client_name(session: Session, client_name: str) -> List[SiteSQL]:
    """Get sites from client name.

    :param session: database session
    :param client_name: client name
    :return: list of site objects
    """
    logger.debug(f"Getting {client_name}'s sites")

    # start main query
    query = session.query(SiteSQL)

    # join the Client table
    query = query.join(ClientSQL)

    # order by uuuid
    query = query.order_by(SiteSQL.site_uuid)

    # select the sites related to the client name
    query = query.filter(ClientSQL.client_name == client_name)

    # get all results
    sites: Optional[List[SiteSQL]] = query.all()

    if len(sites) == 0:
        raise Exception(f"Could not find sites from client {client_name}")

    return sites
